#!/usr/bin/env python3
import os
import time
from datetime import datetime
from typing import List, Tuple, Dict

import click
import psutil
import pyarrow as pa
import pyarrow.parquet as pq
import pandas as pd
from rich.console import Console
from rich.table import Table

from ecopass.core.record_header import RecordHeader


# -------------------------------------------------
# Custom Display to show memory usage
# -------------------------------------------------
def get_memory_usage() -> float:
    """Get current memory usage in MB."""
    process = psutil.Process(os.getpid())
    return process.memory_info().rss / 1024 / 1024


# -------------------------------------------------------------------------------------------------------------
# Custom Display Function To show smooth animation of loading with ram usage and record counting in real-time
# -------------------------------------------------------------------------------------------------------------
def display_utils(batch_count: int, records: List[Tuple[int, bytes]], total_records: int) -> None:
    spinner_frames = ['⠋', '⠙', '⠹', '⠸', '⠼', '⠴', '⠦', '⠧', '⠇', '⠏']  # Spinner frames
    frame = spinner_frames[batch_count % len(spinner_frames)]
    message = (
        f"{frame} Processing batch {batch_count}: {len(records):,} records | "
        f"Total: {total_records:,} | "
        f"Memory: {get_memory_usage():.2f} MB | "
        f"Time: {datetime.now().strftime('%H:%M:%S')}"
    )
    click.echo(f"\r{message}", nl=False)


# -------------------------------------------------------------------------------
# Custom Time Format in second will output in terminal using click custom message
# -------------------------------------------------------------------------------
def format_time(seconds: float) -> str:
    """Format time duration in a human-readable format."""
    hours = seconds // 3600
    minutes = (seconds % 3600) // 60
    seconds = seconds % 60
    return f"{int(hours)}h {int(minutes)}m {int(seconds)}s"



def validate_header_size(ctx, param, value):
    if value not in {2, 4}:
        raise click.BadParameter('header-size must be 2 or 4.')
    return value
# ---------------------------------------------------------------------------------------------------------------------------------------
# User Input: Define File Paths Output_Dir with header-size and alignment and also custom batch-size to control the ram and cpu if needed
# ---------------------------------------------------------------------------------------------------------------------------------------
@click.command(context_settings=dict(help_option_names=['-h', '--help']), cls=click.Command, help='Process binary files with specified options.')
@click.option(
    '--input', 'path_input', required=True,
    type=click.Path(exists=True, dir_okay=False, readable=True),
    help="Path to the binary file to process."
)
@click.option(
    '--output', 'path_output', required=True,
    type=click.Path(file_okay=False, writable=True),
    help="Output directory for encoded data as Parquet files."
)
@click.option(
    '--batch-size', default=500_000, show_default=True,
    type=click.IntRange(min=1),
    help="Batch size (number of records per batch)."
)
@click.option(
    '--header-size', default=2, show_default=True,
    type=click.INT,
    callback=validate_header_size,
    help=(
        "Length of the file header. The first 4 bits are always set to 3 (0011 in binary) "
        "indicating that this is a system record. The remaining bits contain the length of "
        "the file header record, unless the file is an indexed file type that does not include "
        "a separate index file (such as IDXFORMAT8); for those types of file, see Index File - "
        "File Header. If the maximum record length is less than 4095 bytes, the length is 126 "
        "and is held in the next 12 bits; otherwise it is 124 and is held in the next 28 bits. "
        "Hence, in a file where the maximum record length is less than 4095 bytes, this field "
        "contains x'30 7E 00 00'. Otherwise, this field contains x'30 00 00 7C'."
    )
)
@click.option(
    '--alignment', default=2, show_default=True,
    type=click.IntRange(min=2),
    help="Alignment to use with RecordHeader."
)
@click.option(
    '--debug', default=False, show_default=True,
    type=bool,
    help="Enable debug mode to display detailed logs and error messages for troubleshooting."
)
def main(path_input: str, path_output: str, batch_size: int, header_size: int, alignment: int, debug: bool) -> None:
    # -------------------------------------------
    # program start at start_time
    # -------------------------------------------
    start_time = time.time()

    # -------------------------------------------
    # Define Schema for pyarrow & pandas DataFrame
    # -------------------------------------------
    schema = pa.schema([
        ('rdw', pa.int32()),
        ('value', pa.binary())
    ])

    # ------------------------------------------------------------------------------------------
    # Read Records in Batches
    # ------------------------------------------------------------------------------------------
    # # batch size based on memory constraints -> 1_900_000 * 836 bytes = ~950 MB
    # ------------------------------------------------------------------------------------------
    # batch_size = 10_000
    # ------------------------------------------------------------------------------------------
    
    records = []
    total_records = 0
    batch_count = 0

    
    click.echo(f"\n{'='*50}")
    click.echo(f"Processing started at: {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}")
    click.echo(f"Batch size: {batch_size:,} records")
    click.echo(f"{'='*50}\n")

    # ------------------------------------------------------------------------------------------
    # with RecordHeader(filename=path_input, header_size=2, alignment=2) as record_header:
    # ------------------------------------------------------------------------------------------
    with RecordHeader(filename=path_input, header_size=header_size, alignment=alignment, debug=debug) as record_header:
        for length, data in record_header.read_records():
            records.append((length, data))
            total_records += 1
            
            # ---------------------------------------------
            # Write to Parquet when batch size is reached
            # ---------------------------------------------
            if len(records) >= batch_size:
                batch_count += 1

                # ------------------------------------------------
                # Console Ouput using Custom DIsplay using click
                # ------------------------------------------------
                display_utils(batch_count=batch_count, records=records, total_records=total_records)


                df = pd.DataFrame(records, columns=['rdw', 'value'])

                # ---------------------------------------------
                # Convert DataFrame to PyArrow Table
                # ---------------------------------------------
                table = pa.Table.from_pandas(df, schema=schema)
                # table = pa.Table.from_arrays([pa.array([length for length, _ in records]),pa.array([data for _, data in records])], schema=schema)
                # ---------------------------------------------
                # Write to Parquet file
                # ---------------------------------------------
                pq.write_to_dataset(table, root_path=path_output, partition_cols=['rdw'])

                # ---------------------------------------------
                # Clear records to free memory
                # ---------------------------------------------
                records.clear()

    # ---------------------------------------------
    # Write remaining records if any
    # ---------------------------------------------
    if records:
        batch_count += 1
        display_utils(batch_count=batch_count, records=records, total_records=total_records)
        df = pd.DataFrame(records, columns=['rdw', 'value'])
        table = pa.Table.from_pandas(df, schema=schema)
        pq.write_to_dataset(table, root_path=path_output, partition_cols=['rdw'])
        records.clear()
    # ---------------------------------------------
    # since the app convert into cli this will show the ouput of the program
    # Clear Jupyter Notebook Output (if applicable)
    # ---------------------------------------------
    # Final statistics
    # ---------------------------------------------
    end_time = time.time()
    duration = end_time - start_time

    # ---------------------------------------------
    # Initialize the console
    # ---------------------------------------------
    click.echo("\n")
    console = Console()

    # ---------------------------------------------
    # Create a table
    # ---------------------------------------------
    table = Table(title="Processing Summary")

    # ---------------------------------------------
    # Add columns
    # ---------------------------------------------
    table.add_column("Metric", style="cyan", no_wrap=True)
    table.add_column("Value", style="magenta")



    # ---------------------------------------------
    # Add rows
    # ---------------------------------------------
    table.add_row("[bold]Total records processed[/bold]", f"{total_records:,}")
    table.add_row("[bold]Number of [red]batches[/red][/bold]", f"{batch_count}")
    table.add_row("[bold]Average records per [red]batch[/red][/bold]", f"{total_records/batch_count:,.0f}")
    table.add_row("[bold]Total processing time[/bold]", format_time(duration))
    table.add_row("[bold]Processing rate[/bold]", f"{total_records/duration:,.0f} [red]R/S[/red]")
    table.add_row("[bold]Final memory usage[/bold]", f"{get_memory_usage():.2f} MB")
    table.add_row("[bold]Completed at[/bold]", datetime.now().strftime('%Y-%m-%d %H:%M:%S'))

    # ---------------------------------------------
    # Display the table
    # ---------------------------------------------
    console.print(table)

if __name__ == '__main__':
    main()


"""
This module contains the RecordHeader class, which is used to read the header of a record from a file sequentially.
It is used to iterate over the records of a file.

The header is 2 or 4 bytes long.
The first byte is the status and the second byte is the length.
The status is the first 4 bits of the first byte.
The length is the remaining 12 bits of the first byte in case of 2 bytes header.
The length is the remaining 28 bits of the first byte in case of 4 bytes header.

The status is used to determine the type of the record.
The length is used to determine the length of the record.

The record is the data after the header.
The record is aligned to the alignment.

The RecordHeader class is used to read the header of a record and return the length and the data.
It is used to iterate over the records of a file.
"""
import enum
import mmap
from typing import Tuple

class RecordType(enum.Enum):
    SYSTEM_RECORD_DUPLICATE = 1  # A system record (duplicate occurrence)
    DELETED_RECORD = 2  # Deleted record (available for reuse)
    SYSTEM_RESERVED = 3  # System record
    NORMAL_USER_DATA = 4  # Normal user data record
    REDUCED_USER_DATA = 5  # Reduced user data record (indexed files only)
    POINTER_RECORD = 6  # Pointer record (indexed files only)
    USER_DATA_RECORD = 7  # User data record referenced by a pointer
    REDUCED_USER_DATA_REFERENCED_BY_POINTER = 8  # Reduced user data record referenced by a pointer
    RESERVED_FOR_FUTURE_USE = 9  # Reserved for future use
    MID_TRANSACTION_USER_RECORD = 10  # Mid-transaction user data record
    MID_TRANSACTION_REDUCED_USER_RECORD = 11  # Mid-transaction reduced user data record
    MID_TRANSACTION_USER_RECORD_REFERENCED = 12  # Mid-transaction user record referenced by a pointer
    MID_TRANSACTION_REDUCED_USER_RECORD_REFERENCED = 13  # Mid-transaction reduced user record referenced by a pointer


class RecordHeader():
    """
    RecordHeader is a class that reads the header of a record.
    It is used to read the header of a record and return the length and the data.
    It is used to iterate over the records of a file.
    """
    def __init__(self, filename, record_length: list = [], header_size: int = 2, alignment: int = 2, debug=False):
        """
        Initialize the RecordHeader class.
        """
        if header_size not in {2, 4}:
            raise ValueError("Only 2-byte or 4-byte headers are supported")
        
        self.filename = filename
        self.header_size = header_size
        self.record_length = record_length
        # -------------------------------------------------------------------------------------------------------------
        # for now we will display all record since we only discover just two types [system reserved] & [normal user data]
        # __noraml_user_data__ & __system reserved__
        # -------------------------------------------------------------------------------------------------------------
        self.allowed_statuses = [
            RecordType.NORMAL_USER_DATA.value,
            RecordType.DELETED_RECORD.value,
            RecordType.SYSTEM_RESERVED.value,
            RecordType.USER_DATA_RECORD.value,
            RecordType.RESERVED_FOR_FUTURE_USE.value,
            RecordType.SYSTEM_RECORD_DUPLICATE.value,
            RecordType.POINTER_RECORD.value,
            RecordType.REDUCED_USER_DATA.value,
            RecordType.MID_TRANSACTION_USER_RECORD.value,
            RecordType.MID_TRANSACTION_REDUCED_USER_RECORD.value,
            RecordType.MID_TRANSACTION_USER_RECORD_REFERENCED.value,
            RecordType.MID_TRANSACTION_REDUCED_USER_RECORD_REFERENCED.value,
        ]
        self.alignment = alignment
        self.file = None        # Will be set in __enter__
        self.mmap_obj = None    # Will be set in __enter__
        self.debug = debug      # Debug mode flag __track_errors & __fix_bugs__

    #--------------------------------------------
    # open the file and create a memory map for it
    # when the context is __enter__ the file is opened and the memory map is created
    #--------------------------------------------
    def __enter__(self):
        """Opens the file and creates a memory map for it."""
        # --------------------------------------------      
        # open the file
        # --------------------------------------------
        self.file = open(self.filename, 'rb')
        # --------------------------------------------
        # create a memory map for the file
        # --------------------------------------------
        self.mmap_obj = mmap.mmap(self.file.fileno(), length=0, access=mmap.ACCESS_READ)
        return self

    #--------------------------------------------
    # close the memory map and file when exiting the context
    #--------------------------------------------
    def __exit__(self, exc_type, exc_value, traceback):
        """
        Closes the memory map and file when exiting the context.
        """
        self._close()

    #--------------------------------------------
    # parse the header
    #--------------------------------------------
    def _parse_header(self, header_bytes: bytes) -> Tuple[int, int]:
        """
        Parse the header bytes into a status and length.
        The header is 2 or 4 bytes long.
        The first byte is the status and the second byte is the length.
        - The status is the first 4 bits of the first byte.
        - The length is the remaining 12 bits of the first byte in case of 2 bytes header.
        - The length is the remaining 28 bits of the first byte in case of 4 bytes header.
        """
        if self.header_size == 2:
            byte0, byte1 = header_bytes
            # Extract status (first 4 bits of byte0)
            status = (byte0 >> 4) & 0x0F
            # Extract length (remaining 12 bits)
            length = ((byte0 & 0x0F) << 8) | byte1
        elif self.header_size == 4:
            # Extract status (first 4 bits of byte0)
            byte0, byte1, byte2, byte3 = header_bytes
            status = (byte0 >> 4) & 0x0F
            # Extract length (remaining 12 bits)
            length = ((byte0 & 0x0F) << 24) | (byte1 << 16) | (byte2 << 8) | byte3

        return status, length
    
    #--------------------------------------------
    # Calculate padding
    #--------------------------------------------
    def _parse_padding(self, length: int):
        """
        Calculate the padding size.
        The padding size is the difference between the total header data and the alignment.
        """ 
        # --------------------------------------------
        # Example :
        # (2 - (107 + 2 % 2)) % 2
        # --------------------------------------------
        # if length % 2 != 0:
        #     f.seek(1, 1)
        # --------------------------------------------
        total_header_data = self.header_size + length
        padding_size = (self.alignment - (total_header_data % self.alignment)) % self.alignment
        return padding_size
    
    #--------------------------------------------
    # close the file and the memory map
    #--------------------------------------------
    def _close(self):
        """
        Close the file and the memory map.
        """
        if self.mmap_obj:
            self.mmap_obj.close()
        if self.file:
            self.file.close()

    #--------------------------------------------
    # read the records
    #--------------------------------------------
    def read_records(self):
        """
        Read the records from the memory map.
        - Read the header
        - Read the data
        - Calculate the padding
        - Seek to the next record
        - Return the length and the data
        """
        try:
            f = self.mmap_obj
            while True:
                #-------------------------------------------------------------------
                # Read the 2-byte or 4-byte header
                #-------------------------------------------------------------------
                pos = f.tell()
                header = f.read(self.header_size)

                #-------------------------------------------------------------------
                # if the header is not complete, break
                # when the header is not complete, the header is not a valid header
                # reached the end of the file
                #-------------------------------------------------------------------
                if len(header) < self.header_size:
                    if self.debug:
                        print("--------------------------------")
                        print(f"remaining bytes: {f.read(100).hex()}")
                        print(f"header len: {len(header)}, expected header size: {self.header_size}")
                        print(f"\n[DEBUG] Reading header at position {pos}: bytes={header.hex()}")
                        print("[DEBUG] Reached end of file (incomplete header)")
                        print("--------------------------------")
                    break  # End of file

                status, length = self._parse_header(header)

                #-------------------------------------------------------------------
                # if the status is not valid, break
                #-------------------------------------------------------------------
                if status not in self.allowed_statuses:
                    if self.debug:
                        print(f"remaining bytes: {f.read(self.header_size)}")
                        print("--------------------------------")
                        print(f"[DEBUG] Invalid status: {status}")
                        print("--------------------------------")
                
                #-------------------------------------------------------------------------------------
                # used as a example to stop the iteration
                #-------------------------------------------------------------------------------------
                # if header == b'@k' or header == b'@L':
                #     break
                
                #-------------------------------------------------------------------------------------
                # enable debug mode only for simple test since will print a lot of data in the console
                #-------------------------------------------------------------------------------------
                if status in self.allowed_statuses:
                    if self.debug:
                        print("--------------------------------")
                        print(f"[DEBUG] Parsed -> status={status}, length={length}")
                        print(f"[DEBUG] Record type: {RecordType(status).name}")
                        print(f"[DEBUG] Expected data length: {length} bytes")
                        print("--------------------------------")

                #-------------------------------------------------------------------------------------  
                # Read the record data
                #-------------------------------------------------------------------------------------
                data = f.read(length)
                actual_data_length = len(data)
                
                #-------------------------------------------------------------------------------------
                # if the data is not complete, raise an error
                #-------------------------------------------------------------------------------------
                if actual_data_length < length:
                    self._close()
                    if self.debug:
                        print("\n--------------------------------")
                        print(f"[DEBUG] ERROR: Incomplete data at {pos}")
                        print(f"[DEBUG] Expected {length} bytes, got {actual_data_length}")
                        print("\n--------------------------------")
                    raise ValueError(f"Incomplete data at position {pos}")
                
                #-------------------------------------------------------------------------------------
                # return each iterator
                #-------------------------------------------------------------------------------------
                # if status in self.allowed_statuses and length in self.record_length:
                #-------------------------------------------------------------------------------------
                if status in self.allowed_statuses:
                    yield length, bytes(data)

                #-------------------------------------------------------------------------------------
                # calculate the padding size
                #-------------------------------------------------------------------------------------
                padding_size = self._parse_padding(length=length)
                f.seek(padding_size, 1)
        except:
            #-------------------------------------------------------------------------------------
            # close the file and raise an error
            # stop the iteration
            #-------------------------------------------------------------------------------------
            self._close()
            raise ValueError(f"Incomplete data at position {pos}")
        finally:
            ...
